package goblincave.gitea.nes;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.RandomAccessFile;
import java.net.URISyntaxException;
import java.util.LinkedHashMap;
import java.util.Set;

import org.apache.commons.io.FilenameUtils;

import com.diogonunes.jcolor.Ansi;
import com.diogonunes.jcolor.Attribute;

/**
 * Ridge Racer 6 Audio Extractor
 * @author Nes
 * @version 1.0, 2024-07-14
 */
public class AudioExtractor {
	
	/**
	 * Activate alternate behavior for command line terminal.
	 * Specifically prints things in a narrower width, and overwrites previous lines to add color.
	 */
	private final static boolean CMD_MODE = true;
	
	public static void test(String[] args) {
		
		int maxVal = Integer.MAX_VALUE;
		System.out.println(maxVal);
		
		byte[] bytes = Analyzer.intToLittleEndian(maxVal);
		System.out.printf("0x%02x%02x%02x%02x\n", bytes[0], bytes[1], bytes[2], bytes[3]);
		
		maxVal = Analyzer.littleEndianToInt(bytes);
		System.out.println(maxVal);
		
		bytes = Analyzer.intToLittleEndian(maxVal);
		System.out.printf("0x%02x%02x%02x%02x\n", bytes[0], bytes[1], bytes[2], bytes[3]);
		
	}
	
	/**
	 * BIN files must first be extracted from game disk to use this program.
	 * By default uses program directory for I/O and saves audio as WAV.<br>
	 * 
	 * You can simply run this program in the same directory as <b>pack_*.bin</b> files
	 * to launch with default arguments: <pre>extract currentDir currentDir wav</pre>
	 * 
	 * @param args - 4 optional arguments<ol>
	 * <li>"extract" or "pack" mode</li>
	 * <li>package directory containing <b>pack_*.bin</b> files</li>
	 * <li>extract directory for unpacked audio files</li>
	 * <li>extract BGM in "WAV" or "FLAC" format</li></ol>
	 * @throws URISyntaxException 
	 * @throws InterruptedException 
	 */
	public static void main(String[] args) throws URISyntaxException, InterruptedException {
		
		if(args.length > 0) {
			switch(args[0].toLowerCase()) {
				case "identify": {
					File file = new File(args[0]);
					identify(file);
					break;
				}
				case "extract": {
					File packageFile = new File(args[0]);
					extract(packageFile);
					break;
				}
				case "convert": {
					if(args.length < 2)
						throw new IllegalArgumentException("Missing one or more required parameters.");
					
				}
				default: throw new IllegalArgumentException(args[0] + " is not a valid command.");
			}
		} else {
			
			// Print Audio Extractor logo
			if(CMD_MODE) {
				
				// NARROW LOGO
				System.out.println(Ansi.colorize(Logo.format("Ridge", true, 8), Attribute.BRIGHT_RED_TEXT()));
				Thread.sleep(500);
				
				String adjustment = "█████████   ████████████████████████████████████████████████████                     ████████████████████████";
				System.out.println(ANSI_BACKLINE + ANSI_BACKLINE + Ansi.colorize(adjustment, Attribute.BRIGHT_RED_TEXT()));
				System.out.println(Ansi.colorize(Logo.format("Racer 6", true, 9), Attribute.BRIGHT_RED_TEXT()));
				Thread.sleep(500);
				
				System.out.println(Ansi.colorize(Logo.format("Audio", true, 8), Attribute.YELLOW_TEXT()));
				Thread.sleep(500);
				
				System.out.println(Ansi.colorize(Logo.format("Xtractr", true, 8), Attribute.YELLOW_TEXT()));
				Thread.sleep(500);
				
			} else {
				
				// WIDE LOGO
				System.out.println(Ansi.colorize(Logo.format("Ridge Racer 6"), Attribute.BRIGHT_RED_TEXT()));
				Thread.sleep(500);
				
				System.out.println(Ansi.colorize(Logo.format("Audio Extractor", true, 8), Attribute.YELLOW_TEXT()));
				Thread.sleep(500);
				
			}
			
			// Print meta info
			System.out.println("Program:\tRidge Racer 6 Audio Extractor " + Ansi.colorize("v1.0", Attribute.BRIGHT_GREEN_TEXT()) + " by " 
					+ Ansi.colorize("Nes", Attribute.TEXT_COLOR(252, 42, 124)));
			System.out.println("Repository:\t" + Ansi.colorize("https://gitea.goblincave.synology.me/Nes/RR6AudioExtractor", Attribute.CYAN_TEXT()));
			System.out.println();
			
			// Initialize user input reader
			BufferedReader reader = new BufferedReader(new InputStreamReader(System.in));
			
			while(true) {
			
				// determine which operation to execute
				String operationInput = null;
				if(args.length > 0)
					operationInput = args[0];
				String operation = retrieveOperation(reader, operationInput);
				
				if(args.length == 0) { // no arguments provided
								
					switch(operation) {
						
						case "identify": {
							File unknownFile = retrieveFile(reader, "Please enter the path of a file: " + ANSI_BEIGE);
							identify(unknownFile);
							break;
						}
						
						case "extract": {
							File packageFile = retrievePackage(reader);
							if(packageFile.isFile())
								extract(packageFile);
							else { // packageFile is a folder
								File[] packages = packageFile.listFiles(new FilenameFilter() {
									public boolean accept(File dir, String name) {
										return name.toLowerCase().endsWith(".bin");
									}
								});
								for(File pack : packages) {
									extract(pack);
								}
							}
							break;
						}
						
						case "convert": {
							File audioFile = retrieveAudio(reader); // Confirms file has WAV/XMA/FLAC extension + header
							//state which encoding file uses
							//user selects target format
							String format = retrieveFormat(reader);
//							if(audioFile.isFile()) {
//								String extension = FilenameUtils.getExtension(audioFile.getName());
//								// Ensure file has the right encoding (differentiate between WAV/XMA)
//								int encoding = 0;
//								RandomAccessFile raf;
//								try {
//									raf = new RandomAccessFile(audioFile, "r");
//									raf.seek(0x0);
//									raf.readInt();
//								} catch (FileNotFoundException e) {
//									// TODO Auto-generated catch block
//									e.printStackTrace();
//								}
								
								convert(audioFile, format);
//							}
							
						}
						
						case "package": {// pack
							break;
						}
						case "patch": // patch
							File waveFile = retrieveWave(reader);
							patch(waveFile);
							break;
						case "print": // print
							Attribute color = retrieveColor(reader);
							String message = retrieveInput(reader, "Please enter your logo message: " + ANSI_BEIGE);
							print(color, message);
							break;
						case "exit": // exit
							System.out.println("Program closed.");
							System.exit(0);
							break;
					}
		
				}
				
				System.out.println("Press Enter to continue...");
				try { reader.readLine(); } catch (Exception e) {}
			
			}
		}

	}

	/**
	 * Helper function to prompt user to select a program task.
	 * 
	 * @param reader
	 * @param input
	 * @return operation
	 */
	private static String retrieveOperation(BufferedReader reader, String input) {
		
		String operation = null;
		String[] values = {"identify", "extract", "convert", "patch", "pack", "print", "exit"};
		
		if(input != null)
			for(int i = 0; i < values.length; i++)
				if(values[i].equalsIgnoreCase(input))
					operation = values[i];
		
		if(operation == null) {
		
			String prompt = Ansi.colorize("Operations", Attribute.BRIGHT_YELLOW_TEXT(), Attribute.BOLD()) + ":\n"
					+ "- " + Ansi.colorize("Identify", Attribute.BRIGHT_RED_TEXT()) + " a package or audio file\n"
					+ "- " + Ansi.colorize("Extract", Attribute.TEXT_COLOR(0xFF, 0x68, 0x1F)) + " audio tracks from a package (BIN -> WAV/XMA)\n"
					+ "- " + Ansi.colorize("Convert", Attribute.TEXT_COLOR(0xFF, 0xEA, 0x00)) + " audio tracks to another format (XMA <-> WAV <-> FLAC)\n"
					+ "- " + Ansi.colorize("Patch", Attribute.BRIGHT_GREEN_TEXT()) + " an audio track to repeat (WAV)\n"
					+ "- " + Ansi.colorize("Pack", Attribute.BRIGHT_BLUE_TEXT()) + " audio tracks into a package (WAV/XMA -> BIN)\n"
					+ "- " + Ansi.colorize("Print", Attribute.TEXT_COLOR(0x6B, 0x4C, 0xE3)) + " a Ridge Racer 6 style ASCII logo\n"
					+ "- " + Ansi.colorize("Exit", Attribute.BRIGHT_MAGENTA_TEXT()) + " program\n"
					+ "\nPlease select an operation: " + ANSI_BEIGE;
			
			operation = retrieveInput(reader, prompt, values);
			
			if(CMD_MODE) {
				
				Attribute color = null;
				switch(operation.toLowerCase()) {
					case "identify":
						color = Attribute.BRIGHT_RED_TEXT();
						break;
					case "extract":
						color = Attribute.TEXT_COLOR(0xFF, 0x68, 0x1F);
						break;
					case "convert":
						color = Attribute.TEXT_COLOR(0xFF, 0xEA, 0x00);
					case "patch":
						color = Attribute.BRIGHT_GREEN_TEXT();
						break;
					case "pack":
						color = Attribute.BRIGHT_BLUE_TEXT();
						break;
					case "print":
						color = Attribute.TEXT_COLOR(0x6B, 0x4C, 0xE3);
						break;
					case "exit":
						color = Attribute.BRIGHT_MAGENTA_TEXT();
						break;
				}
				System.out.println(ANSI_BACKLINE + Ansi.colorize("Please select an operation: ") + Ansi.colorize(operation, color));
				
			} else System.out.print(Ansi.colorize(""));
		
		}
		
		return operation.toLowerCase();
		
	}
	
	/**
	 * Prompts the user for input and validates that the input matches one of the provided values.
	 * 
	 * @param reader
	 * @param prompt
	 * @param values
	 * @return input
	 */
	private static String retrieveInput(BufferedReader reader, String prompt, String[] values) {
		
		String input = null;
		
		boolean valid = false;
		do {
			System.out.print(prompt);
			try {
				input = reader.readLine();
				for(int i = 0; i < values.length; i++)
					if(input.equalsIgnoreCase(values[i])) {
						valid = true;
						break;
					}
				if(!valid)
					System.out.println(Ansi.colorize("Invalid selection."));
			} catch (IOException e) {
				System.out.println(Ansi.colorize("Invalid selection."));
			}
		} while(!valid);
		
		return input;
		
	}

	/**
	 * Extracts audio tracks from the given package file.
	 * 
	 * @param packageFile
	 * @throws FileNotFoundException 
	 */
	public static void extract(File packageFile) {
		
		System.out.println("Name:\t" + packageFile.getName());
		System.out.println("Size:\t" + packageFile.length() + " (" + Analyzer.convertBytes(packageFile.length()) + ")\n");
		
		try {
			RandomAccessFile file = new RandomAccessFile(packageFile, "r");
			LinkedHashMap<Integer, Integer> tracklist = Analyzer.findAudioTracks(file);
			
			File dir = new File(packageFile.getAbsoluteFile().getParentFile().toPath() + "/" 
					+ FilenameUtils.removeExtension(packageFile.getName()));
			if(!dir.exists())
				dir.mkdir();
			
			System.out.println("Extracting " + tracklist.size() + " tracks...");

			int digits = (int) Math.ceil(Math.log10(tracklist.size() + 1));
			int extracted = 0;
			Set<Integer> keys = tracklist.keySet();
			for(Integer address : keys) {
				int length = tracklist.get(address);
				
				if(CMD_MODE)
					printProgressBar(extracted, keys.size());
				
				String extension;
				byte[] bytes = new byte[2];
				file.seek(address + 0x14);
				file.read(bytes);
				int encoding = Analyzer.littleEndianToInt(bytes);
				switch(encoding) {
					case 0x0165: extension = "xma"; break;
					case 0x0001: extension = "wav"; break;
					default: extension = "bin"; break;
				}
				
				File trackFile = new File(dir.toPath() + "/track" + String.format("%0" + digits + "d", extracted + 1) + "." + extension);
				
				try (
						FileInputStream in = new FileInputStream(packageFile); 
						FileOutputStream out = new FileOutputStream(trackFile);
				) {
					in.skipNBytes(address.intValue());
					bytes = new byte[length];
					int readBytes = in.read(bytes);
					out.write(bytes, 0, readBytes);
				} catch (IOException e) {
					e.printStackTrace();
				}
				extracted++;
			}
			
			if(CMD_MODE)
				printProgressBar(extracted, keys.size());
			
		} catch(Exception e) {
			e.printStackTrace();
		}
		
		System.out.println("Finished.\n");
		
	}

//	/**
//	 * Extracts audio files to the extract directory from package files located at the package directory.
//	 * Also supports FLAC conversion for BGM files.
//	 * 
//	 * @param packDirectory
//	 * @param extractDirectory
//	 * @param compressBGM
//	 * @deprecated delete after adapting FLAC conversion
//	 */
//	public static void extract(File packDirectory, File extractDirectory, boolean compressBGM) {
//		
//		System.out.println("packDirectory:\t" + packDirectory);
//		System.out.println("extractDirectory:\t" + extractDirectory);
//		
//		// Identify target binary files
//		File[] packages = new File(packDirectory.toString()).listFiles(new FilenameFilter() {
//			public boolean accept(File dir, String name) {
//				return name.toLowerCase().endsWith(".bin");
//			}
//		});
//		
//		// Identify and extract audio within binary files
//		for(int i = 0; i < packages.length; i++) {
//			
//			RandomAccessFile source;
//			LinkedHashMap<Integer, Integer> tracklist = null;
//			try {
//				source = new RandomAccessFile(packages[i], "r");
//				tracklist = Analyzer.findAudioTracks(source);
//			} catch (FileNotFoundException e) {
//				System.out.println("Binary file:\t" + packages[i] + "\n was not found. File skipped.");
//				continue;
//			} catch (IOException e) {
//				System.out.println("Binary file:\t" + packages[i] + "\n could not be read. File skipped.");
//				continue;
//			}
//			
//			System.out.println("packages[" + i + "]:\t" + packages[i]);
//					
//			File dir = new File(extractDirectory.getPath() + "/" + FilenameUtils.removeExtension(packages[i].getName()));
//			if(!dir.exists())
//				dir.mkdir();
//			
//			// Extract and write WAV files in directory
//			Set<Integer> keys = tracklist.keySet();
//			int track = 1;
//			
//			for(Integer key : keys) {
//				
//				String name = String.format("track%04d", track);
//				{ // Extract WAV files
//					File wav = null;
//					try {
//							
//						// create file for storing WAV data
//						wav = new File(dir.getPath() + "/" + name + ".wav");
//						if(!wav.exists())
//							wav.createNewFile();
//					
//						System.out.println("Saving track " + String.format("%d (@0x%08X)", track, key) + " to " + wav + String.format(" (%d bytes)", tracklist.get(key)));
//						
//						// write selection to file
//						try (
//								FileInputStream inStream = new FileInputStream(packages[i]); 
//								FileOutputStream outStream = new FileOutputStream(wav)
//						) {
//							inStream.skipNBytes(key.intValue());
//							byte[] trackBytes = new byte[tracklist.get(key)];
//							int readBytes = inStream.read(trackBytes);
//							outStream.write(trackBytes, 0, readBytes);
//						} catch (IOException e) {
//							e.printStackTrace();
//							System.exit(3);
//						}
//						
//					} catch (IOException e) {
//						System.out.println("An error occurred when attempting to write " + name + " to file:\t" + wav);
//						e.printStackTrace();
//						System.exit(3);
//					}
//				}
//				
//				track++;
//				
//			}
//			
//			if(compressBGM && packages[i].getName().equals("pack_bgm.bin")) {
//
//				// find WAV files
//				File[] WAV_Files = dir.listFiles(new FilenameFilter() {
//					public boolean accept(File dir, String name) {
//						return name.toLowerCase().endsWith(".wav");
//					}
//				});
//				
//				// convert WAV files to FLAC
//				FLAC_FileEncoder ffe = new FLAC_FileEncoder();
//								
//				for(File wav : WAV_Files) {
//				
//					try {
//						File flac = new File(dir.getPath() + "/" + FilenameUtils.removeExtension(wav.getName()) + ".flac");
//						if(!flac.exists())
//							flac.createNewFile();
//						
//						System.out.println("Compressing WAV to FLAC:\t" + flac);
//					
//						ffe.encode(wav, flac);
//					} catch (IOException e) {
//						System.out.println("An error occurred when attempting to write file.");
//						e.printStackTrace();
//						System.exit(3);
//					}
//					
//				}
//				
//				System.out.print("Deleting WAV files");
//				// delete WAV files
//				for(File wav : WAV_Files) {
//					boolean deleted = false;
//					while(!deleted) {
//						try {
//							Files.delete(wav.toPath());
//							deleted = true;
//						} catch(Exception e) {
//							System.out.print('.');
//						}
//					}
//				}
//				
//			}
//			
//		}
//		
//	}
	
	/**
	 * Helper method to prompt user to select a package file, or folder containing one or more packages.
	 * 
	 * @param reader
	 * @return file
	 */
	private static File retrievePackage(BufferedReader reader) {
		
		String prompt = "Please enter the path of a package: " + ANSI_BEIGE;
		
		File packageFile = null;
		boolean valid = false;
		do {
			
			// Retrieve a file or folder
			packageFile = retrieveFileOrFolder(reader, prompt);
			
			if(packageFile.isFile()) {
				// File must have .bin extension
				String name = packageFile.getName();
				int extIndex = name.lastIndexOf('.');
				String extension = null;
				if(extIndex > 0 && extIndex < name.length() - 1)
					extension = name.substring(extIndex + 1);
				if(extension != null && extension.equalsIgnoreCase("bin"))
					valid = true;
			} else { // packageFile is folder
				; // Folder should have a .bin file
				File[] contents = packageFile.listFiles();
				for(File file : contents) {
					String name = file.getName();
					int extIndex = name.lastIndexOf('.');
					String extension = null;
					if(extIndex > 0 && extIndex < name.length() - 1)
						extension = name.substring(extIndex + 1);
					if(extension != null && extension.equalsIgnoreCase("bin")) {
						valid = true;
						break;
					}
				}
			}
			
			if(!valid) {
				if(packageFile.isFile())
					System.out.println("File does not have a .BIN file extension.");
				else System.out.println("Directory does not contain a .BIN file.");
			}
			
		} while(!valid);
		
		return packageFile;
		
	}
	
	/**
	 * Helper function to prompt user to select a file or folder.
	 * 
	 * @param reader
	 * @return file or folder
	 */
	private static File retrieveFileOrFolder(BufferedReader reader, String prompt) {
		
		File file = null;
		
		String input = null;
		boolean valid = false;
		do {
			System.out.print(prompt);
			try {
				input = reader.readLine();
				if(input == null)
					throw new IOException("Invalid file.");
				file = new File(input);
				if(file.exists()) {
					valid = true;
				} else System.out.println(Ansi.colorize("File does not exist."));
			} catch (IOException e) {
				System.out.println(Ansi.colorize("Invalid file."));
			}
		} while(!valid);
		
		System.out.print(Ansi.colorize("\n"));
		
		return file;
		
	}
	
	/**
	 * Helper function to show current progress.
	 * Should only be used in CMD Mode.
	 * 
	 * @param current
	 * @param total
	 */
	private static void printProgressBar(int current, int total) {

		int width = 50; // progress bar total width
		double progress = (double) current / total; // 0 = none, 1 = complete
		
		int progressWidth = (int) Math.floor(progress * width);
		double progressRemainder = (progress * width) % 1.0;
		int x = (int) Math.floor(progressRemainder * 8);
		char[] c = { ' ', '▏', '▎', '▍', '▌', '▋', '▊', '▉' };
		
		String bar = "";
		for(int i = 0; i < width; i++) {
			if(i < progressWidth)
				bar += '█';
			else if(i == progressWidth)
				bar += c[x];
			else bar += ' ';
		}
		
		System.out.println(ANSI_BACKLINE + "Progress: |" + bar + "| (" + current + "/" + total + ")");
	}
	
	/**
	 * Examines a given file and reads some of its properties.
	 * 
	 * @param givenFile
	 */
	public static void identify(File givenFile) {
		
		String name = givenFile.getName();
		System.out.println("Name:\t" + name);
		
		long size = givenFile.length();
		System.out.println("Size:\t" + size + " Bytes" + (size > 1024 ? " (" + Analyzer.convertBytes(size) + ")" : "") + "\n");
		
		// file extension
		int extIndex = name.lastIndexOf('.');
		String extension = null;
		if(extIndex > 0 && extIndex < name.length() - 1)
			extension = name.substring(extIndex + 1);
		
		try {
			RandomAccessFile file = new RandomAccessFile(givenFile, "r");
			switch(extension.toLowerCase()) {
				case "bin": {
					LinkedHashMap<Integer, Integer> tracklist = Analyzer.findAudioTracks(file);
					System.out.println(tracklist.size() + " Tracks (" + Analyzer.convertBytes(size / tracklist.size()) + " avg size)\n");
					System.out.println("Format Info");
					Analyzer.readFormatChunk(file, 0x0C);
					break;
				}
				case "flac": {
					
					break;
				}
				case "wav":
				case "xma":
				default: {
					for(long i = 0; i < size - 4; i += 4) {
						file.seek(i);
						int value = file.readInt();
						switch(value) {
							case Analyzer.ASCII_RIFF: {
								Analyzer.readRIFFChunk(file, (int) i);
								break;
							}
							case Analyzer.ASCII_fmt: {
								i = Analyzer.readFormatChunk(file, (int) i) - 4;
								break;
							}
							case Analyzer.ASCII_smpl: {
								i = Analyzer.readSampleChunk(file, (int) i) - 4;
								break;
							}
							case Analyzer.ASCII_data: {
								i = Analyzer.readDataChunk(file, (int) i) - 4;
								break;
							}
							case Analyzer.ASCII_ALIG: {
								i = Analyzer.readAlignmentChunk(file, (int) i) - 4;
								break;
							}
							case Analyzer.ASCII_seek: {
								i = Analyzer.readx2stChunk(file, (int) i) - 4;
								break;
							}
							case Analyzer.ASCII_x2st: {
								i = Analyzer.readx2stChunk(file, (int) i) - 4;
								break;
							}
							case Analyzer.ASCII_LIST: {
								i = Analyzer.readListChunk(file, (int) i) - 4;
								break;
							}
						}
					}
					break;
				}
			}
		} catch (IOException e) {
			System.out.println("Could not read file.");
		}
		
	}
	
	/**
	 * Helper function to prompt user to select a file.
	 * 
	 * @param reader
	 * @return file
	 */
	private static File retrieveFile(BufferedReader reader, String prompt) {
		
		File file = null;
		
		String input = null;
		boolean valid = false;
		do {
			System.out.print(prompt);
			try {
				input = reader.readLine();
				if(input == null)
					throw new IOException("Invalid file.");
				file = new File(input);
				if(file.exists()) {
					if(file.isFile()) {
						valid = true;
					} else System.out.println(Ansi.colorize("Path does not resolve to a file."));
				} else System.out.println(Ansi.colorize("File does not exist."));
			} catch (IOException e) {
				System.out.println(Ansi.colorize("Invalid file."));
			}
		} while(!valid);
		
		System.out.print(Ansi.colorize("\n"));
		
		return file;
		
	}
	
	/**
	 * Helper function to prompt user to select a file.
	 * 
	 * @param reader
	 * @return file
	 */
	private static File retrieveFile(BufferedReader reader, String prompt, String extension) {
		
		File file = null;
		
		String input = null;
		boolean valid = false;
		do {
			System.out.print(prompt);
			try {
				input = reader.readLine();
				file = new File(input);
				if(file.exists()) {
					if(file.getName().toLowerCase().endsWith(extension)) {
						valid = true;
					} else System.out.println(Ansi.colorize("File does not have " + extension.toUpperCase() + " file extension"));
				} else System.out.println(Ansi.colorize("File does not exist."));
			} catch (IOException e) {
				System.out.println(Ansi.colorize("Invalid file."));
			}
		} while(!valid);
		
		return file;
		
	}
	
	/**
	 * Converts an audio file to the target format.
	 *  
	 * @param audioFile
	 * @param format
	 */
	public static void convert(File audioFile, String format) {
		
		System.out.println("convert(" + audioFile.toPath() + ", " + format + ") not yet implemented.");
		
	}

	/**
	 * Helper function to prompt user to select a convertible audio file.
	 * 
	 * @param reader
	 * @return audio file
	 */
	private static File retrieveAudio(BufferedReader reader) {
		
		File file = null;
		
		String input = null;
		boolean valid = false;
		do {
			System.out.print("Please enter the path of an audio file:\t" + ANSI_BEIGE);
			try {
				input = reader.readLine();
				file = new File(input);
				if(file.exists()) {
					if(file.isFile()) {
						
						String name = file.getName().toLowerCase();
						String extension = FilenameUtils.getExtension(name);
						
						if(extension.equals("wav") || extension.equals("xma")) {
							// check for RIFF header
							byte[] bytes = new byte[4];
							try (RandomAccessFile raf = new RandomAccessFile(file, "r")) {
								raf.seek(0x0);
								if(raf.readInt() == Analyzer.ASCII_RIFF) {
									raf.seek(0x14);
									raf.read(bytes);
									int encoding = Analyzer.littleEndianToInt(bytes);
									if(encoding == 0x0001)
										valid = true;
									else if(encoding == 0x0165)
										valid = true;
									else {
										System.out.println("Unsupported encoding" + Analyzer.identifyEncoding(encoding) + ".");
									}
								}
								else System.out.println("Not a valid RIFF file.");
							}
							
						} else if (name.endsWith(".flac")) {
							// check for FLAC header
							try (RandomAccessFile raf = new RandomAccessFile(file, "r")) {
								raf.seek(0);
								if(raf.readInt() == Analyzer.ASCII_fLaC)
									valid = true;
								else System.out.println("Not a valid FLAC file.");
							}
							
						} else System.out.println(Ansi.colorize("File does not have a supported file extension."));
						
					}
				} else System.out.println(Ansi.colorize("File does not exist."));
			} catch (IOException e) {
				System.out.println(Ansi.colorize("Invalid file."));
			}
		} while(!valid);
		
		return file;
		
	}
	
	/**
	 * Prompts user to select an audio format that can be converted to.
	 * 
	 * @param reader
	 * @return
	 */
	private static String retrieveFormat(BufferedReader reader) {
		
		String prompt = "\nSupported audio formats: "
		+ "\n- " + Ansi.colorize("WAV", Attribute.BRIGHT_BLUE_TEXT()) 				+ "\tPCM audio (uncompressed)"
		+ "\n- " + Ansi.colorize("FLAC", Attribute.TEXT_COLOR(0xFF, 0x68, 0x1F)) 	+ "\tPCM audio (lossless compression)"
		+ "\n- " + Ansi.colorize("XMA", Attribute.BRIGHT_GREEN_TEXT()) 				+ "\tXbox Media Audio (lossy compression)"
		+ "\n" 
		+ "\n" + Ansi.colorize("Please select an audio format:") + ANSI_BEIGE + " ";
		
		String[] values = {"WAV", "XMA", "FLAC"};
		
		String input = retrieveInput(reader, prompt, values);
		
		Attribute color = null;
		switch(input.toLowerCase()) {
			case "wav":		color = Attribute.BRIGHT_BLUE_TEXT(); 			break;
			case "flac":	color = Attribute.TEXT_COLOR(0xFF, 0x68, 0x1F);	break;
			case "xma": 	color = Attribute.BRIGHT_GREEN_TEXT(); 			break;
		}
		if(CMD_MODE)
			System.out.println(ANSI_BACKLINE + Ansi.colorize("Please select a color: ") + Ansi.colorize(input, color) + "\n");
		else System.out.println(Ansi.colorize(""));
		
		return input;
		
	}
	
	/**
	 * Patches a WAV file by seeking out its data and smpl chunks, 
	 * replacing any existing smpl chunk with a new smpl chunk,
	 * setting loop points at the start and end of the data chunk.
	 * 
	 * @param waveFile
	 */
	public static void patch(File waveFile) {
		
		try {
			RandomAccessFile file = new RandomAccessFile(waveFile, "r");
			
			// Must find 3 chunks to successfully patch WAV file:
			int fmt_Address = -1;
			int dataAddress = -1;
			int smplAddress = -1;
			
			int fmt_Size = -1;
			int dataSize = -1;
			int smplSize = -1;
			
			int sampleWidth = -1;
			
			boolean[] found = {false, false, false};
			for(int i = 0; i < file.length(); i += 4) {
				
				file.seek(i);
				int scan = file.readInt();
				
				// System.out.println(i + "\t" + scan);
				
				// Skip to end of chunk when found				
				if(scan == Analyzer.ASCII_fmt) {
					System.out.println("format chunk found");
					found[0] = true;
					fmt_Address = i;
					fmt_Size = Analyzer.readChunkSize(file, i);
					
					// get sample width
					byte[] bytes = new byte[2];
					file.seek(i + 0x14); 
					file.read(bytes);
					sampleWidth = Analyzer.littleEndianToInt(bytes);
					
					i += fmt_Size - 4;
					System.out.println("Skipping to " + i);
					
				} else if(scan == Analyzer.ASCII_data) {
					System.out.println("data chunk was found");
					found[1] = true;
					dataAddress = i;
					dataSize = Analyzer.readChunkSize(file, i);
					i += dataSize - 4;
					
				} else if(scan == Analyzer.ASCII_smpl) {
					
					found[2] = true;
					smplAddress = i;
					smplSize = Analyzer.readChunkSize(file, i);
					i += smplSize - 4;
					
				}
				
				// if all chunks already located, stop seeking
				if(found[0] && found[1] && found[2])
					break;
				
			}
			
			System.out.println("chunks searched");
			
			if(!found[0]) {
				System.out.println("format chunk missing");
				throw new IOException("Format information not found in file.");
			}
			if(!found[1]) {
				System.out.println("data chunk missing");
				throw new IOException("Audio data not found in file.");
			}
			
			byte[] smpl = new byte[68];
			// smpl (0x00~3)
			smpl[0x00] = 0x73; // s
			smpl[0x01] = 0x6D; // m
			smpl[0x02] = 0x70; // p
			smpl[0x03] = 0x6C; // l
			// size (0x04~7)
			smpl[0x04] = 0x3C; // 60 bytes
			// manufacturer (0x08~B)
			// product (0x0C~F)
			// sample period (0x10~3)
			// MIDI unity note (0x14~7)
			smpl[0x14] = 0x3C; // pitch 60
			// MIDI pitch fraction (0x18~B)
			// SMPTE Format (0x1C~F)
			// SMPTE Offset (0x20~3)
			// Number of Sample Loops (0x24~7)
			smpl[0x24] = 0x01; // 1 loop
			// Sample Loops size (0x28~B)
			// loop ID (0x2C~F)
			// loop type (0x30~3)
			// loop start (0x34~7)
			smpl[0x34] = 0x01; // start at sample 0
			// loop end (0x38~B)
			byte[] loopEnd = Analyzer.intToLittleEndian((dataSize - 8) / sampleWidth - 1);
			smpl[0x38] = loopEnd[0]; // data end
			smpl[0x39] = loopEnd[1];
			smpl[0x3A] = loopEnd[2];
			smpl[0x3B] = loopEnd[3];
			// tuning fraction (0x3C~F)
			// play count (0x40~3)
			
			System.out.println("chunks found");
			
			if(!found[2]) {
				File dir = waveFile.getAbsoluteFile().getParentFile();
				File patched = new File(dir + "/" + FilenameUtils.removeExtension(waveFile.getName()) + "_patched.wav");
				if(!patched.exists())
					patched.createNewFile();
				try (
						FileInputStream in = new FileInputStream(waveFile);
						FileOutputStream out = new FileOutputStream(patched);
				) {
					
					int totalSize = (int) waveFile.length() + smpl.length - 8; 
					byte[] sizeBytes = Analyzer.intToLittleEndian(totalSize);
					
					byte[] bytes = new byte[dataAddress];
					int readBytes = in.read(bytes);
					
					bytes[4] = sizeBytes[0];
					bytes[5] = sizeBytes[1];
					bytes[6] = sizeBytes[2];
					bytes[7] = sizeBytes[3];
					
					out.write(bytes, 0, readBytes); 						// file up to start of data chunk
					
					out.write(smpl, 0, smpl.length);				// sample chunk
					
					int part2Size = (int) waveFile.length() - dataAddress;
					bytes = new byte[part2Size];
					readBytes = in.read(bytes);
					out.write(bytes, 0, readBytes); // data chunk and beyond
					
				}
			// Create new sample chunk
				// Insert 68-byte sample chunk with loop before data chunk
			
			// Set loop points to start and end of data chunk
			// Update the new file size in RIFF header
			}
			
			System.out.println("patched");
			
		} catch(IOException e) {
			System.out.println("Failed to patch file.");
		}
		
	}
	
	/**
	 * Helper function to prompt user to select a WAV file.
	 * 
	 * @param reader
	 * @return WAV file
	 */
	private static File retrieveWave(BufferedReader reader) {
		
		String prompt = "Please provide a WAV file: " + ANSI_BEIGE;
		
		File waveFile = null;
		boolean valid = false;
		do {
			
			// Retrieve an audio file
			waveFile = retrieveFile(reader, prompt, ".wav");
			
			try {
				// validate that chosen file is a proper Wave file
				RandomAccessFile file = new RandomAccessFile(waveFile, "r");
				// first 4 bytes start with RIFF
				if(file.readInt() == Analyzer.ASCII_RIFF) {
					// next 4 bytes contain the size of the file in little-endian representation
					int headerSize = Analyzer.readChunkSize(file, 0);
//					file.seek(4);
//					byte[] buffer = new byte[4];
//					file.read(buffer); // write size bytes to buffer
//					
//					// calculate expected file size
//					int headerSize = 8 + buffer[3] 
//							+ buffer[2] * (int) Math.pow(16, 2) 
//							+ buffer[1] * (int) Math.pow(16, 4)
//							+ buffer[0] * (int) Math.pow(16, 6);
					
					// file is valid if actual file length matches expected size
					valid = headerSize == file.length();
					if(!valid)
						System.out.println("File length mismatch. Expected " + headerSize + " bytes, but found " + file.length() + " bytes.");
					
				} else System.out.println("Invalid RIFF header.");
				file.close();
			} catch (IOException e) {
				System.out.println("Invalid file.");
			}
			
		} while(!valid);
		
		return waveFile;
		
	}
	
	/**
	 * Packs WAV files located at the extract directory into BIN files at the package directory.
	 * 
	 * @param packDirectory
	 * @param extractDirectory
	 */
	public static void pack(File packDirectory, File extractDirectory) {
		
		// Identify extracted folders
		File[] packages = new File(extractDirectory.toString()).listFiles(new FilenameFilter() {
			public boolean accept(File dir, String name) {
				return name.toLowerCase().startsWith("pack_") && !name.toLowerCase().endsWith(".bin");
			}
		}); 
        
		for(int i = 0; i < packages.length; i++)
			System.out.println(packages[i].getName());
		
		for(File packDir : packages) {
			
			// 1. Compile list of WAV files in extracted directories
			File[] audioTracks = new File(packDir.toString()).listFiles(new FilenameFilter() {
				public boolean accept(File dir, String name) {
					return name.toLowerCase().endsWith(".wav");
				}
			});
			
			// 2. Check if pack files already exist in package directory
			File pack = new File(packDir.getPath() + ".bin");
			if(pack.exists());
			
			// 3. If pack file already exists, insert WAV files into existing file.
			// 4. If pack file doesn't exist, create new file then insert WAV files.
			
		}
		
		// Hypothesis 1: Does Ridge Racer 6 use a predetermined address list for reading tracks from package?
		// Test 1: Try changing the address of a track and see if the game can handle it successfully.
		
		// Hypothesis 2: Does Ridge Racer 6 loop tracks in their entirety, or do they use loop points defined in the file header?
		// Test 1: Try editing loop information and see if it affects in-game playback.
		// Test 2: Try replacing a track with a WAV file without loop data and see if it loops in in-game playback.
		
	}

	/**
	 * Prints a RR6 ASCII style logo with the given color and message.
	 * 
	 * @param color
	 * @param message
	 */
	public static void print(Attribute color, String message) {
		
		System.out.println(Ansi.colorize(Logo.format(message), color));
		
	}
	
	/**
	 * Helper function to prompt user to select a color.
	 * 
	 * @param reader
	 * @return color attribute
	 */
	private static Attribute retrieveColor(BufferedReader reader) {
		
		String prompt = // "\nCommand parameters: " + Ansi.colorize("Print", Attribute.BRIGHT_BLUE_TEXT()) 
				//+ Ansi.colorize(" [color]", Attribute.BRIGHT_YELLOW_TEXT()) + Ansi.colorize(" [message]", Attribute.BRIGHT_RED_TEXT()) 
				//+ "\n" 
				/*+*/ "\n" + Ansi.colorize("Colors", Attribute.BRIGHT_YELLOW_TEXT(), Attribute.BOLD()) + ":"
				+ "\n- " + Ansi.colorize("Default", Attribute.CLEAR())
				+ "\n- " + Ansi.colorize("Black", Attribute.BLACK_TEXT()) 
						+ ", " + Ansi.colorize("Gray", Attribute.BRIGHT_BLACK_TEXT()) 
						+ ", " + Ansi.colorize("White", Attribute.BRIGHT_WHITE_TEXT())
				+ "\n- " + Ansi.colorize("Indigo", Attribute.TEXT_COLOR(0x7D, 0x1A, 0xFF)) 
						+ ", " + Ansi.colorize("Purple", Attribute.BRIGHT_MAGENTA_TEXT()) 
						+ ", " + Ansi.colorize("Violet", Attribute.TEXT_COLOR(0xEE, 0x82, 0xEE))
				+ "\n- " + Ansi.colorize("Red", Attribute.RED_TEXT()) 
						+ ", " + Ansi.colorize("Cinnabar", Attribute.BRIGHT_RED_TEXT()) 
						+ ", " + Ansi.colorize("Pink", Attribute.TEXT_COLOR(0xFF, 0xC0, 0xCB))
				+ "\n- " + Ansi.colorize("Orange", Attribute.TEXT_COLOR(0xFF, 0x68, 0x1F)) 
						+ ", " + Ansi.colorize("Gold", Attribute.YELLOW_TEXT()) 
						+ ", " + Ansi.colorize("Yellow", Attribute.TEXT_COLOR(0xFF, 0xEA, 0x00))
				+ "\n- " + Ansi.colorize("Green", Attribute.BRIGHT_GREEN_TEXT()) 
						+ ", " + Ansi.colorize("Chartreuse", Attribute.TEXT_COLOR(0xD1, 0xEB, 0x27)) 
						+ ", " + Ansi.colorize("Beige", Attribute.BRIGHT_YELLOW_TEXT())
				+ "\n- " + Ansi.colorize("Blue", Attribute.BRIGHT_BLUE_TEXT()) 
						+ ", " + Ansi.colorize("Teal", Attribute.TEXT_COLOR(0x00, 0x93, 0xCF)) 
						+ ", " + Ansi.colorize("Cyan", Attribute.BRIGHT_CYAN_TEXT())
				+ "\n" 
				+ "\n" + Ansi.colorize("Please select a color:") + ANSI_BEIGE + " ";
		String[] values = {"Default", "Black", "Gray", "White",	"Indigo", "Purple", "Violet", "Red", "Cinnabar", "Pink", "Orange", "Gold", 
				"Yellow", "Green", "Chartreuse", "Beige","Blue", "Teal", "Cyan"};
		
		String input = retrieveInput(reader, prompt, values);
		
		Attribute color = null;
		switch(input.toLowerCase()) {
			case "default":		color = Attribute.CLEAR(); 						break;
			case "black": 		color = Attribute.BLACK_TEXT(); 				break;
			case "gray": 		color = Attribute.BRIGHT_BLACK_TEXT(); 			break;
			case "white": 		color = Attribute.BRIGHT_WHITE_TEXT(); 			break;
			case "indigo": 		color = Attribute.TEXT_COLOR(0x7D, 0x1A, 0xFF); break;
			case "purple": 		color = Attribute.BRIGHT_MAGENTA_TEXT(); 		break;
			case "violet": 		color = Attribute.TEXT_COLOR(0xEE, 0x82, 0xEE); break;
			case "red": 		color = Attribute.RED_TEXT(); 					break;
			case "cinnabar": 	color = Attribute.BRIGHT_RED_TEXT(); 			break;
			case "pink": 		color = Attribute.TEXT_COLOR(0xFF, 0xC0, 0xCB); break;
			case "orange":		color = Attribute.TEXT_COLOR(0xFF, 0x68, 0x1F);	break;
			case "gold":		color = Attribute.YELLOW_TEXT();				break;
			case "yellow":		color = Attribute.TEXT_COLOR(0xFF, 0xEA, 0x00);	break;
			case "green": 		color = Attribute.BRIGHT_GREEN_TEXT(); 			break;
			case "chartreuse": 	color = Attribute.TEXT_COLOR(0xD1, 0xEB, 0x27); break;
			case "beige":		color = Attribute.BRIGHT_YELLOW_TEXT(); 		break;
			case "blue":		color = Attribute.BRIGHT_BLUE_TEXT(); 			break;
			case "teal":		color = Attribute.TEXT_COLOR(0x00, 0x93, 0xCF); break;
			case "cyan":		color = Attribute.BRIGHT_CYAN_TEXT(); 			break;
		}
		
		if(CMD_MODE)
			System.out.println(ANSI_BACKLINE + Ansi.colorize("Please select a color: ") + Ansi.colorize(input, color) + "\n");
		else System.out.println(Ansi.colorize(""));
		
		return color;
		
	}

	/**
	 * Prompts the user for input without strict validation.
	 * 
	 * @param reader
	 * @param prompt
	 * @return input
	 */
	private static String retrieveInput(BufferedReader reader, String prompt) {
		
		String input = null;
		
		boolean valid = false;
		do {
			System.out.print(prompt);
			try {
				input = reader.readLine();
				valid = true;
			} catch (IOException e) {
				System.out.println(Ansi.colorize("Invalid input."));
			}
		} while(!valid);
		
		return input;
		
	}

	/**
	 * ANSI code to go up one line, followed by a carriage return.
	 * Useful for overwriting user input.
	 */
	private final static String ANSI_BACKLINE = "\033[F\r";
	/**
	 * ANSI code to switch to beige colored text.
	 * Useful for highlighting user input.
	 */
	private final static String ANSI_BEIGE = "\u001B[93m";
	
	
}
